-- Import global variables

local assert, error, ipairs, next, pairs, select, tonumber, tostring, type, unpack, pcall, xpcall =
      assert, error, ipairs, next, pairs, select, tonumber, tostring, type, unpack, pcall, xpcall
local getmetatable, setmetatable, rawequal, rawget, rawset, getfenv, setfenv, loadstring, debugstack =
      getmetatable, setmetatable, rawequal, rawget, rawset, getfenv, setfenv, loadstring, debugstack
local math, string, table = math, string, table
local find, format, gmatch, gsub, tolower, match, toupper, join, split, trim =
      string.find, string.format, string.gmatch, string.gsub, string.lower, string.match, string.upper, string.join, string.split, string.trim
local concat, insert, maxn, remove, sort = table.concat, table.insert, table.maxn, table.remove, table.sort
local max, min, abs, ceil, floor = math.max, math.min, math.abs, math.ceil, math.floor
local _G = getfenv()

local LibStub = LibStub

local GetTime = GetTime
local IsMouselooking = IsMouselooking
local MouselookStart = MouselookStart
local MouselookStop = MouselookStop
local SetMouselookOverrideBinding = SetMouselookOverrideBinding
local SpellIsTargeting = SpellIsTargeting

-- No global variables after this!


-- ================================================================================================
-- = Localization                                                                                 =
-- ================================================================================================

local L = LibStub("AceLocale-3.0"):NewLocale("MouseLook", "enUS", true)
L["General"] = true
L["Triggers"] = true
L["Mouse buttons"] = true
L["Mouse button control"] = true
L["Toggle mouse-look on/off"] = true
L["Turn mouse-look on/off momentarily"] = true
L["Turn mouse-look on"] = true
L["Turn mouse-look off"] = true
L["Auto-run"] = true
L["Enter mouse-look mode when you begin auto-running."] = true
L["Track movement"] = true
L["Keep track of whether you are moving or not. This can improve auto-run handling since it is not possible to determine if you are currently auto-running. NOTE: It is not possible to track movement inside an instance!"] = true
L["Auto-follow"] = true
L["Leave mouse-look mode when you begin auto-following a unit."] = true
L["Right-click"] = true
L["Enter mouse-look mode when holding down the right mouse button for a short amount of time."] = true
L["Right-click delay"] = true
L["Set the amount of time you need to hold down the right mouse button before entering mouse-look mode."] = true
L["Spell-cast"] = true
L["Leave mouse-look mode when you cast a spell that needs a mouse-click to target."] = true
L["Left-click override"] = true
L["Adjust behaviour of the left mouse-button when you are in mouse-look mode."] = true
L["Right-click override"] = true
L["Adjust behaviour of the right mouse-button when you are in mouse-look mode."] = true
L["Do nothing"] = true

local L = LibStub("AceLocale-3.0"):GetLocale("MouseLook")

-- ================================================================================================
-- =  Initialization (addon, libraries, etc.)                                                     =
-- ================================================================================================
_G.BINDING_HEADER_MOUSELOOK = "MouseLook"
_G.BINDING_NAME_MOUSELOOK_TOGGLE = L["Toggle mouse-look on/off"]
_G.BINDING_NAME_MOUSELOOK_ON = L["Turn mouse-look on"]
_G.BINDING_NAME_MOUSELOOK_OFF = L["Turn mouse-look off"]
_G.BINDING_NAME_MOUSELOOK_HOLD = L["Turn mouse-look on/off momentarily"]

local Core = LibStub("AceAddon-3.0"):NewAddon("MouseLook", "AceConsole-3.0", "AceEvent-3.0", "AceHook-3.0")

local MinRightClickDelay = .1
local MaxRightClickDelay = 1
local DefaultRightClickDelay = .7

local function FixDelay(delay)
  return delay and max(MinRightClickDelay, min(MaxRightClickDelay, floor(delay * 100 + .5) / 100)) or DefaultRightClickDelay
end

local ClickOverrides = {}
do
  ClickOverrides.NONE = L["Do nothing"]
  local bindings = {
    "MOUSELOOK_OFF", "STRAFELEFT", "STRAFERIGHT", "MOVEFORWARD", "MOVEBACKWARD",
    "JUMP", "SITORSTAND", "TOGGLEAUTORUN", "TOGGLERUN",
  }
  for _, binding in ipairs(bindings) do
    ClickOverrides[binding] = _G["BINDING_NAME_"..binding] or binding
  end
end

local function FixClickOverride(click_override)
  if click_override and ClickOverrides[click_override] then return click_override end
  return "NONE"
end

local EnableMoveTracker, DisableMoveTracker, IsMoving
do
  local moving, lastx, lasty
  function IsMoving() return moving end
  local MoveTracker
  function EnableMoveTracker()
    if not MoveTracker then
      local GetPlayerMapPosition = _G.GetPlayerMapPosition
      MoveTracker = _G.CreateFrame("Frame")
      local timer = 0
      MoveTracker:SetScript("OnUpdate", function(self, elapsed)
        if timer <= elapsed then
          timer = 0.2
          local x, y = GetPlayerMapPosition("player")
          if not lastx then lastx, lasty = x, y end
          if not moving and (lastx ~= x or lasty ~= y) then
            moving = true
          elseif moving and lastx == x and lasty == y then
            moving = nil
          end
          lastx, lasty = x, y
        else
          timer = timer - elapsed
        end
      end)
      MoveTracker:SetScript("OnEvent", function()
        _G.SetMapToCurrentZone()
      end)
    else
      MoveTracker:Show()
    end
    MoveTracker:RegisterEvent("ZONE_CHANGED_NEW_AREA")
  end
  function DisableMoveTracker()
    if MoveTracker then 
      MoveTracker:Hide()
      MoveTracker:UnregisterAllEvents()
      moving, lastx, lasty = nil
    end
  end
end

local GetOptions

function Core:OnInitialize()
  local defaults_profile = {
    autorun = true,
    movetracking = true,
    autofollow = true,
    rightclick = true,
    rightclick_delay = DefaultRightClickDelay,
    leftclick_override = "NONE",
    rightclick_override = "MOUSELOOK_OFF",
    spellcast = true,
  }

  local defaults = {profile = defaults_profile}

  self.db = LibStub("AceDB-3.0"):New("MouseLookAce3DB", defaults, "Default")
  self.db.RegisterCallback(self, "OnProfileChanged", "OnProfileChanged")
  self.db.RegisterCallback(self, "OnProfileCopied", "OnProfileChanged")
  self.db.RegisterCallback(self, "OnProfileReset", "OnProfileChanged")

  LibStub("AceConfigRegistry-3.0"):RegisterOptionsTable("MouseLook", GetOptions)
  self:RegisterChatCommand("mouselook", "OpenConfig", true)
  self:RegisterChatCommand("ml", "OpenConfig", true, false)
end

function Core:OpenConfig()
  self:MouselookOff()
  LibStub("AceConfigDialog-3.0"):Open("MouseLook")
end

function Core:OnEnable()
  self:OnProfileChanged()
end

function Core:OnDisable()
  SetMouselookOverrideBinding("BUTTON1", nil)
  SetMouselookOverrideBinding("BUTTON2", nil)
  self:MouselookOff()
  DisableMoveTracker()
end

function Core:IsAutoRun()
  return not not self.db.profile.autorun
end

function Core:SetAutoRun(enable)
  self.db.profile.autorun = not not enable
  self:EnableHooksAndEvents()
end

function Core:IsMoveTracking()
  return not not self.db.profile.movetracking
end

function Core:SetMoveTracking(enable)
  self.db.profile.movetracking = not not enable
  self:EnableHooksAndEvents()
end

function Core:IsAutoFollow()
  return not not self.db.profile.autofollow
end

function Core:SetAutoFollow(enable)
  self.db.profile.autofollow = not not enable
  self:EnableHooksAndEvents()
end

function Core:IsRightClick()
  return not not self.db.profile.rightclick
end

function Core:SetRightClick(enable)
  self.db.profile.rightclick = not not enable
  self:EnableHooksAndEvents()
end

function Core:GetRightClickDelay()
  return self.db.profile.rightclick_delay
end

function Core:SetRightClickDelay(delay)
  self.db.profile.rightclick_delay = FixDelay(delay)
end

function Core:GetLeftClickOverride()
  return self.db.profile.leftclick_override
end

function Core:SetLeftClickOverride(key)
  self.db.profile.leftclick_override = key
  self:EnableHooksAndEvents()
end

function Core:GetRightClickOverride()
  return self.db.profile.rightclick_override
end

function Core:SetRightClickOverride(key)
  self.db.profile.rightclick_override = key
  self:EnableHooksAndEvents()
end

function Core:IsSpellCast()
  return not not self.db.profile.spellcast
end

function Core:SetSpellCast(enable)
  self.db.profile.spellcast = not not enable
  self:EnableHooksAndEvents()
end

function Core:SecureHookOrUnhook(enable, ...)
  local hooked = self:IsHooked((...))
  if enable then
    if not hooked then
      self:SecureHook(...)
    end
  elseif hooked then
    self:Unhook((...))
  end
end

function Core:EnableHooksAndEvents()
  if not self:IsEnabled() then return end
  local autorun = self:IsAutoRun()
  self:SecureHookOrUnhook(autorun, "ToggleAutoRun")
  if autorun and self:IsMoveTracking() then
    EnableMoveTracker()
  else
    DisableMoveTracker()
  end

  local rightclick = self:IsRightClick()
  self:SecureHookOrUnhook(rightclick, "TurnOrActionStart")
  self:SecureHookOrUnhook(rightclick, "TurnOrActionStop")

  if self:IsSpellCast() then
    self:RegisterEvent("CURSOR_UPDATE")
  else
    self:UnregisterEvent("CURSOR_UPDATE")
  end

  if self:IsAutoFollow() then
    self:RegisterEvent("AUTOFOLLOW_BEGIN")
  else
    self:UnregisterEvent("AUTOFOLLOW_BEGIN")
  end

  SetMouselookOverrideBinding("BUTTON1", self:GetLeftClickOverride() or "NONE")
  SetMouselookOverrideBinding("BUTTON2", self:GetRightClickOverride() or "NONE")
end

function Core:OnProfileChanged()
  self.db.profile.rightclick_delay = FixDelay(self.db.profile.rightclick_delay)

  self.db.profile.leftclick_override = FixClickOverride(self.db.profile.leftclick_override)
  self.db.profile.rightclick_override = FixClickOverride(self.db.profile.rightclick_override)

  self:EnableHooksAndEvents()
end

function Core:ToggleMouselook()
  if IsMouselooking() then MouselookStop() else MouselookStart() end
end

function Core:MouselookOn()
  if not IsMouselooking() then MouselookStart() end
end

function Core:MouselookOff()
  if IsMouselooking() then MouselookStop() end
end

function Core:ToggleAutoRun()
  if IsMoving() then return end
  self:MouselookOn()
end

do
  local time

  function Core:TurnOrActionStart()
    time = GetTime()
  end

  function Core:TurnOrActionStop()
    if not time then return end
    local dt = GetTime() - time
    time = nil
    if not IsMouselooking() and dt >= self:GetRightClickDelay() then
      MouselookStart()
    end
  end
end


function Core:CURSOR_UPDATE()
  if IsMouselooking() and SpellIsTargeting() then
    MouselookStop()
  end
end

function Core:AUTOFOLLOW_BEGIN()
  self:MouselookOff()
end

local options
function GetOptions()
  if options then return options end

  local self = Core

  local _order = 0
  local function order()
    _order = _order + 1
    return _order
  end

  local _spaceCount = 0
  local function AddSpacer(args)
    _spaceCount = _spaceCount + 1
    args["spacer".._spaceCount] = {
      order = order(),
      type = "description",
      name = "",
    }
  end

  options = {
    type = "group",
    childGroups = "tree",
    name = self.name,
    args = {},
  }

  local general = {}
  options.args.general = {
    order = order(),
    name = L["General"],
    type = "group",
    args = general,
  }

  general.triggers = {
    order = order(),
    type = "group",
    name = L["Triggers"],
    inline = true,
    args = {},
  }
  general.triggers.args.autorun = {
    order = order(),
    type = "toggle",
    name = L["Auto-run"],
    desc = L["Enter mouse-look mode when you begin auto-running."],
    get = function() return self:IsAutoRun() end,
    set = function(info, ...) self:SetAutoRun(...) end,
  }
  general.triggers.args.movetracking = {
    order = order(),
    type = "toggle",
    name = L["Track movement"],
    desc = L["Keep track of whether you are moving or not. This can improve auto-run handling since it is not possible to determine if you are currently auto-running. NOTE: It is not possible to track movement inside an instance!"],
    get = function() return self:IsMoveTracking() end,
    set = function(info, ...) self:SetMoveTracking(...) end,
    disabled = function() return not self:IsAutoRun() end,
  }
  AddSpacer(general.triggers.args)
  general.triggers.args.autofollow = {
    order = order(),
    type = "toggle",
    name = L["Auto-follow"],
    desc = L["Leave mouse-look mode when you begin auto-following a unit."],
    get = function() return self:IsAutoFollow() end,
    set = function(info, ...) self:SetAutoFollow(...) end,
  }
  general.triggers.args.spellcast = {
    order = order(),
    type = "toggle",
    name = L["Spell-cast"],
    desc = L["Leave mouse-look mode when you cast a spell that needs a mouse-click to target."],
    get = function() return self:IsSpellCast() end,
    set = function(info, ...) self:SetSpellCast(...) end,
  }
  AddSpacer(general.triggers.args)
  general.triggers.args.rightclick = {
    order = order(),
    type = "toggle",
    name = L["Right-click"],
    desc = L["Enter mouse-look mode when holding down the right mouse button for a short amount of time."],
    get = function() return self:IsRightClick() end,
    set = function(info, ...) self:SetRightClick(...) end,
  }
  general.triggers.args.rightclick_delay = {
    order = order(),
    type = "range",
    name = L["Right-click delay"],
    desc = L["Set the amount of time you need to hold down the right mouse button before entering mouse-look mode."],
    min = MinRightClickDelay,
    max = MaxRightClickDelay,
    get = function() return self:GetRightClickDelay() end,
    set = function(info, ...) self:SetRightClickDelay(...) end,
  }

  general.mousebuttons = {
    order = order(),
    type = "group",
    name = L["Mouse buttons"],
    desc = L["Mouse button control"],
    inline = true,
    args = {},
  }
  general.mousebuttons.args.leftclick_override = {
    order = order(),
    type = "select",
    name = L["Left-click override"],
    desc = L["Adjust behaviour of the left mouse-button when you are in mouse-look mode."],
    get = function() return self:GetLeftClickOverride() end,
    set = function(info, ...) self:SetLeftClickOverride(...) end,
    values = ClickOverrides,
  }
  general.mousebuttons.args.rightclick_override = {
    order = order(),
    type = "select",
    name = L["Right-click override"],
    desc = L["Adjust behaviour of the right mouse-button when you are in mouse-look mode."],
    get = function() return self:GetRightClickOverride() end,
    set = function(info, ...) self:SetRightClickOverride(...) end,
    values = ClickOverrides,
  }

  options.args.profiles = LibStub("AceDBOptions-3.0"):GetOptionsTable(self.db)

  return options
end






